#!/usr/bin/env python

import numpy as np
import pandas as pd
import sklearn
from sklearn.svm import SVC
from sklearn.model_selection import cross_validate
from sklearn.metrics.cluster import adjusted_mutual_info_score
from sklearn.metrics.cluster import adjusted_rand_score
from sklearn.metrics.cluster import homogeneity_score
from sklearn.metrics.cluster import normalized_mutual_info_score


def clustering_metrics(adata, target, pred):   
    """
    Evaluate clustering performance.
    
    Parameters
    ----------
    adata
        AnnData object of shape `n_obs` × `n_vars`. Rows correspond to cells and columns to genes.
    target
        Key in `adata.obs` where ground-truth spatial domain labels are stored.
    pred
        Key in `adata.obs` where clustering assignments are stored.
        
    Returns
    -------
    ami
        Adjusted mutual information score.
    ari
        Adjusted Rand index score.
    homo
        Homogeneity score.
    nmi
        Normalized mutual information score.

    """ 
    ami  = adjusted_mutual_info_score(adata.obs[target], adata.obs[pred])
    ari  = adjusted_rand_score(adata.obs[target], adata.obs[pred])
    homo = homogeneity_score(adata.obs[target], adata.obs[pred])
    nmi  = normalized_mutual_info_score(adata.obs[target], adata.obs[pred])

    return ami, ari, homo, nmi

def mean_average_precision(embedding: np.ndarray, labels: np.ndarray, n_neighbors: int=30, **kwargs) -> float:
    """
    Evaluate mean average precision.

    Parameters
    ----------
    embedding
        Low-dimensional representation of spots.
    labels
        Ground-truth domain labels.
    n_neighbors
        Number of nearest neighbors used to evaluate mean average precision. By default, n_neighbors=30.
    **kwargs
        Additional keyword arguments are passed to
        :class:`sklearn.neighbors.NearestNeighbors`
        
    Returns
    -------
    map
        Mean average precision.
        
    """
    def _average_precision(match: np.ndarray) -> float:
        if np.any(match):
            cummean = np.cumsum(match) / (np.arange(match.size) + 1)
            return cummean[match].mean().item()
        return 0.0
    if isinstance(labels, pd.Series):
        labels = labels.astype(str).values
    knn = sklearn.neighbors.NearestNeighbors(
        n_neighbors=min(labels.shape[0], n_neighbors + 1), **kwargs
    ).fit(embedding)
    nni = knn.kneighbors(embedding, return_distance=False)
    match = np.equal(labels[nni[:, 1:]], np.expand_dims(labels, 1))
    
    return np.apply_along_axis(_average_precision, 1, match).mean().item()

def mean_cross_validation_accuracy(embedding, labels, Kfold=5):
    """
    Evaluate mean cross-validation accuracy.

    Parameters
    ----------
    embedding
        Low-dimensional representation of spots.
    labels
        Ground-truth domain labels.
    Kfold
        Number of folds for cross-validation. By default, Kfold=5.

    Returns
    -------
    MCVA
        Mean cross-validation accuracy.

    """

    if isinstance(labels, pd.Series):
        labels = labels.astype(str).values
    target_unique = np.unique(labels).reshape(1, -1)
    target_onehot = (labels.reshape(-1, 1)==target_unique).astype(int)
    labels = target_onehot.argmax(-1)
    svc = SVC()
    cv_results = cross_validate(svc, embedding, labels, scoring="accuracy", cv=Kfold, n_jobs=Kfold)
    
    return np.mean(cv_results["test_score"])
    


#!/usr/bin/env python

import random
import numpy as np
import scanpy as sc

def setup_seed(seed):
    """
    Set random seed.
    
    Parameters
    ----------
    seed
        Number to be set as random seed for reproducibility.
        
    """
    np.random.seed(seed)
    random.seed(seed)

def get_N_clusters(adata, n_cluster, cluster_method='louvain', range_min=0, range_max=3, max_steps=30, tolerance=0):
    """
    Tune the resolution parameter in clustering to make the number of clusters and the specified number as close as possible.
    
    Parameters
    ----------
    adata
        AnnData object of shape `n_obs` × `n_vars`. Rows correspond to cells and columns to genes.
    n_cluster
        Specified number of clusters.
    cluster_method
        Method (`louvain` or `leiden`) used for clustering. By default, cluster_method='louvain'.
    range_min
        Minimum clustering resolution for the binary search.
    range_max
        Maximum clustering resolution for the binary search.
    max_steps
        Maximum number of steps for the binary search.
    tolerance
        Tolerance of the difference between the number of clusters and the specified number.

    Returns
    -------
    adata
        AnnData object with clustering assignments in `adata.obs`:

        - `adata.obs['louvain']` - Louvain clustering assignments if `cluster_method='louvain'`.
        - `adata.obs['leiden']` - Leiden clustering assignments if `cluster_method='leiden'`.

    """ 
    this_step = 0
    this_min = float(range_min)
    this_max = float(range_max)
    while this_step < max_steps:
        this_resolution = this_min + ((this_max-this_min)/2)
        if cluster_method=='leiden':
            sc.tl.leiden(adata, resolution=this_resolution)
        if cluster_method=='louvain':
            sc.tl.louvain(adata, resolution=this_resolution)
        this_clusters = adata.obs[cluster_method].nunique()

        if this_clusters > n_cluster+tolerance:
            this_max = this_resolution
        elif this_clusters < n_cluster-tolerance:
            this_min = this_resolution
        else:
            print("Succeed to find %d clusters at resolution %.3f."%(n_cluster, this_resolution))
            return adata
        this_step += 1
    
    print('Cannot find the number of clusters.')
    return adata



#!/usr/bin/env python

import numpy as np
import scipy.sparse as sp
from numba import jit
from sklearn.neighbors import kneighbors_graph
from .utils import *

def spatially_variable_genes(adata, mode='moran_geary', n_top_genes=2000, subset=False, inplace=True, 
    coordinate_key='spatial', n_neighbors=30, layer=None, seed=None):
    """
    Identify spatially variable genes by SINFONIA.
    
    Mode
    -----
        * 1. `moran_geary` identify spatially variable genes based on Moran’s I and rescaled Geary’s C scores.
        * 2. `moran` identify spatially variable genes based on Moran’s I scores.
        * 3. `geary` identify spatially variable genes based on rescaled Geary’s C scores.

    Parameters
    ----------
    adata
        AnnData object of shape `n_obs` × `n_vars`. Rows correspond to cells and columns to genes.
    mode
        Mode for identifying spatially variable genes. By default, mode='moran_geary'.
    n_top_genes
        Number of spatially variable genes to keep. By default, n_top_genes=2000.
    subset
        Inplace subset to spatially variable genes if `True` otherwise merely indicate spatially variable genes. By default, subset=False.
    inplace
        Whether to place calculated metrics in `.var`. By default, inplace=True.
    coordinate_key
        Key in `adata.obsm` where spatial coordinates are stored. By default, coordinate_key='spatial'.
    n_neighbors
        Number of nearest neighbors used to construct the spatial neighbor graph. By default, n_neighbors=30.
    layer
        If provided, use `adata.layers[layer]` for expression values instead of `adata.X`. Note that SINFONIA expects logarithmized data. By default, layer=None.
    seed
        Random seed for reproducibility. By default, seed=None.
        
    Returns
    -------
    adata
        AnnData object with identified spatially variable genes and additional fields:

        - `adata.var['spatially_variable']` - boolean indicator of spatially variable genes.
        - `adata.uns['svg']` - mode used to identify spatially variable genes if `inplace=True`.
        - `adata.var['moranI']` - Moran’s I scores if `inplace=True` and `mode='moran_geary'` or `mode='moran'`.
        - `adata.var['gearyC']` - rescaled Geary’s C scores if `inplace=True` and `mode='moran_geary'` or `mode='geary'`.

    """

    if seed is not None: setup_seed(seed)
    sc.pp.filter_genes(adata, min_cells=1)

    @jit(nopython=True, parallel=True)
    def _moran(exp_mtx, data, row, col):
        N = exp_mtx.shape[0]
        sum_W = np.sum(data)
        moran_index = []
        for i in range(exp_mtx.shape[1]):
            exp_vec = exp_mtx[:, i]
            vec_mean = exp_vec.mean()
            C = N*(data*(exp_vec[row]-vec_mean)*(exp_vec[col]-vec_mean)).sum() / (sum_W*np.sum((exp_vec-vec_mean)**2))
            moran_index.append(C)
        return moran_index
    
    @jit(nopython=True, parallel=True)
    def _geary(exp_mtx, data, row, col):
        N = exp_mtx.shape[0]
        sum_W = np.sum(data)
        geary_index = []
        for i in range(exp_mtx.shape[1]):
            exp_vec = exp_mtx[:, i]
            C = (N-1)*(data*((exp_vec[row]-exp_vec[col])**2)).sum() / (2*sum_W*np.sum((exp_vec-exp_vec.mean())**2))
            geary_index.append(1-C)
        return geary_index
    
    @jit(nopython=True, parallel=True)
    def _moran_geary(exp_mtx, data, row, col):
        N = exp_mtx.shape[0]
        sum_W = np.sum(data)
        moran_index = []
        geary_index = []
        for i in range(exp_mtx.shape[1]):
            exp_vec = exp_mtx[:, i]
            vec_mean = exp_vec.mean()
            down_item = np.sum((exp_vec-vec_mean)**2)
            exp_row, exp_col = exp_vec[row], exp_vec[col]
            I = N*(data*(exp_row-vec_mean)*(exp_col-vec_mean)).sum() / (sum_W*down_item)
            C = (N-1)*(data*((exp_row-exp_col)**2)).sum() / (2*sum_W*down_item)
            moran_index.append(I)
            geary_index.append(1-C)
        return moran_index, geary_index
    
    W = kneighbors_graph(adata.obsm[coordinate_key], n_neighbors=n_neighbors,
                        include_self=False, metric="euclidean", mode="distance").tocoo()
    W.data = 1 - W.data/(W.max(axis=1).data.reshape(-1, 1).
                         repeat(n_neighbors, axis=1).reshape(-1) + 1e-5)
    X = adata.layers[layer] if layer is not None else adata.X
    exp_mtx = X.toarray() if sp.issparse(X) else X.copy()
    
    if mode=='moran_geary':
        moran_index, geary_index = _moran_geary(exp_mtx, W.data, W.row, W.col)
        genes_selected1 = np.array([False] * adata.shape[1])
        genes_selected1[np.argpartition(moran_index, len(moran_index)-n_top_genes)[-n_top_genes:]] = True
        genes_selected2 = np.array([False] * adata.shape[1])
        genes_selected2[np.argpartition(geary_index, len(geary_index)-n_top_genes)[-n_top_genes:]] = True    
        genes_selected = np.logical_or(genes_selected1, genes_selected2)
    elif mode=='moran':
        moran_index = _moran(exp_mtx, W.data, W.row, W.col)
        genes_selected = np.array([False] * adata.shape[1])
        genes_selected[np.argpartition(moran_index, len(moran_index)-n_top_genes)[-n_top_genes:]] = True
    elif mode=='geary':    
        geary_index = _geary(exp_mtx, W.data, W.row, W.col)
        genes_selected = np.array([False] * adata.shape[1])
        genes_selected[np.argpartition(geary_index, len(geary_index)-n_top_genes)[-n_top_genes:]] = True
    adata.var['spatially_variable'] = genes_selected
    
    if inplace:
        adata.uns['svg'] = {'mode': mode}
        if mode=='moran_geary':
            adata.var['moranI'] = moran_index
            adata.var['gearyC'] = geary_index
        elif mode=='moran':
            adata.var['moranI'] = moran_index
        elif mode=='geary':    
            adata.var['gearyC'] = geary_index
    if subset:
        adata = adata[:, genes_selected]

    return adata




#!/usr/bin/env python
#-*- coding:utf-8 -*-


from setuptools import setup, find_packages

setup(
    name="sinfonia",
    version="0.0.3",
    keywords=("pip", "sinfonia"),
    description="SINFONIA: scalable identification of spatially variable genes for deciphering spatial domains",
    long_description="SINFONIA provides an effective and efficient way to identify spatially variable genes for deciphering spatial domains. We provide documentation in the form of functional application programming interface documentation, tutorials and example workflows at https://sinfonia-svg.readthedocs.io/en/latest/index.html. All SINFONIA wheels distributed on PyPI are MIT licensed.",
    license="MIT Licence",
    url="https://github.com/BioX-NKU/SINFONIA",
    author="Shengquan Chen",
    packages=find_packages(),
    python_requires='>3.6.0',
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    install_requires=[
        'numpy>=1.21.6,<1.22',
        'pandas>=1.4.2',
        'scipy>=1.8.0',
        'scikit-learn>=1.0.2',
        'numba>=0.55.1',
        'scanpy>=1.9.1'
    ]
)